/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.webservice;

import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;
import org.junit.jupiter.api.Test;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;
import se.digg.dgc.encoding.Base45;
import se.digg.dgc.encoding.Zlib;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.service.impl.DefaultDGCDecoder;
import se.digg.dgc.signatures.CertificateProvider;
import se.digg.dgc.signatures.DGCSignatureVerifier;
import se.digg.dgc.signatures.cose.CoseSign1_Object;
import se.digg.dgc.signatures.cwt.Cwt;
import se.digg.dgc.signatures.impl.DefaultDGCSignatureVerifier;

import java.security.KeyStore;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;

@CbsTest
class CoronaPassWebServiceTest {
    private final ApplicationContext applicationContext = new AnnotationConfigApplicationContext("dk.sundhedsdatastyrelsen.cbs.test");
    private final GetCoronaPassRequest request = new GetCoronaPassRequest();
    private final CoronaPassWebService coronaPassWebService = new CoronaPassWebService(applicationContext);

    private GetCoronaPassRequest request(String cpr) {
        GetCoronaPassRequest request = new GetCoronaPassRequest();
        request.setIdentifier(cpr);
        return request;
    }

    @Test
    void test_getCoronaPass_cache_hit() {
        System.setProperty("test.cache.hit", "true");
        String cprPrefix = "040587010";
        for (int i=0; i<10; i++)
            verifyPass(cprPrefix + i);
    }

    @Test
    void test_getCoronaPass_empty() {
        System.setProperty("test.cache.hit", "true");
        String cpr = "0405870114";
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(0, coronaPass.getCoronaPasses().getPass().size());
    }

    @Test
    void test_getCoronaPassOnlyDK() {
        System.setProperty("test.cache.hit", "true");
        String cpr = "0405870113";
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(2, coronaPass.getCoronaPasses().getPass().size());
    }

    @Test
    void test_getCoronaPassOnlyEU() {
        System.setProperty("test.cache.hit", "true");
        String cpr = "0405870115";
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(3, coronaPass.getCoronaPasses().getPass().size());
    }

    @Test
    void test_getCoronaPass5minutesExpiry() {
        System.setProperty("test.cache.hit", "true");
        String cpr = "0405870118";
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(5, coronaPass.getCoronaPasses().getPass().size());
        PassJWT eu = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).findFirst().get();
        Instant expDate = getEUExp(eu).getExpiration();
        assertTrue(Instant.now().plus(4, ChronoUnit.MINUTES).isBefore(expDate));
        assertTrue(Instant.now().plus(6, ChronoUnit.MINUTES).isAfter(expDate));
    }

    @Test
    void test_verify_signature() throws Exception {
        System.setProperty("test.cache.hit", "true");
        String cpr = "0405870118";
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        PassJWT eu = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).findFirst().get();
        DigitalGreenCertificate certificate = decodeToken(eu.getJWT());
        assertFalse(eu.getProperties().isEmpty());
    }

    @Test
    void test_getCoronaPass15minutesExpiry() {
        System.setProperty("test.cache.hit", "true");
        String cpr = "0405870117";
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        //assertEquals(3, coronaPass.getCoronaPasses().getPass().size());
        PassJWT eu = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).findFirst().get();
        Instant expDate = getEUExp(eu).getExpiration();
        assertTrue(Instant.now().plus(14, ChronoUnit.MINUTES).isBefore(expDate));
        assertTrue(Instant.now().plus(16, ChronoUnit.MINUTES).isAfter(expDate));
    }

    @Test
    void test_getCoronaPass72hourExpiry() {
        System.setProperty("test.cache.hit", "true");
        String cpr = "0405870107";
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(5, coronaPass.getCoronaPasses().getPass().size());
        PassJWT eu = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).findFirst().get();
        Instant expDate = getEUExp(eu).getExpiration();
        assertTrue(Instant.now().plus(71, ChronoUnit.HOURS).isBefore(expDate));
        assertTrue(Instant.now().plus(75, ChronoUnit.HOURS).isAfter(expDate));
    }

    private Cwt getEUExp(PassJWT eu) {
        try {
            byte[] cborBytes = Zlib.decompress(Base45.getDecoder().decode(eu.getJWT().substring(4)), true);
            CoseSign1_Object coseObject = new CoseSign1_Object(cborBytes);
            return coseObject.getCwt();
        } catch (Exception e) {
            throw new IllegalStateException(e);
        }
    }

    public DigitalGreenCertificate decodeToken(String token) throws Exception {
        KeyStore keystore = KeyStore.getInstance("PKCS12");
        keystore.load(this.getClass().getClassLoader().getResourceAsStream("Sundhedsdatastyrelsen - COVID-19 Borgerstatus.p12"), "Test1234".toCharArray());
        Certificate certificate = keystore.getCertificate("covid-19 borgerstatus (funktionscertifikat)");
        DGCSignatureVerifier verifier = new DefaultDGCSignatureVerifier();
        CertificateProvider certificateProvider = (country, kid) -> {
            List<X509Certificate> listOfCertificate = new ArrayList<>();
            listOfCertificate.add((X509Certificate) certificate);
            return listOfCertificate;
        };
        DigitalGreenCertificate decoded = new DefaultDGCDecoder(verifier, certificateProvider).decode(token);
        return decoded;
    }

    private void verifyPass(String cpr) {
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
    }

    @Test
    void test_getCoronaPass_cache_miss() {
        System.setProperty("test.cache.hit", "false");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request);
        assertNotNull(coronaPass);
        assertNotNull(coronaPass.getFutureResult());
        assertNotNull(coronaPass.getFutureResult().getJobId());

        assertNull(coronaPass.getCoronaPasses());
    }
}

